/*
    Title:    DCF77 library
    Author:   Volker Oth
    Date:     5/1999
    Purpose:  Decoding of the DCF77 real time signal.
    needed
    Software: AVR-GCC to compile, AVA to assemble and link
    needed
    Hardware: ATS90S8515 on STK200 board
    Note:     To contact me, mail to
                  volkeroth(at)gmx.de
*/

#include <io.h>
#include <signal.h>
#include "global.h"
#include "timer.h"
#include "dcf77.h"


u08 dcf77_data[DCF77_BUFFER_SIZE];         /* buffer containing all decoded data */
void (*dcf77_call_fall)(void);             /* callback at falling edge */
void (*dcf77_call_rise)(void);             /* callback at rising edge */


static void dcf77_dummy(void)              
/* empty dummmy function for callbacks */
{
}


SIGNAL(SIG_INTERRUPT1)         
/* signal handler for external interrupt int1 */
{
    static u16 fall_time;
    static u16 rise_time;
    register u08 mcucr;

    mcucr = inp(MCUCR);

    if (!(mcucr & (1<<ISC10))) {
        register u16 high_time;

        get_ticks(fall_time);              /* remember time of falling edge */
        if (dcf77_second != 0xff)
            high_time = fall_time-rise_time;
        else
            high_time = 0;

        /* int raised on falling edge */
        mcucr |= (1<<ISC10);               /* raise int on rising edge */

        dcf77_second++;

        if (high_time > DCF77_START) {     /* longer than 1 second high -> startbit */
            if (!(dcf77_status & (1<<DCF77_STATUS_INIT))) {
                /* not yet detected -> startbit detected */
                dcf77_status |=  (1<<DCF77_STATUS_INIT);   
                dcf77_second  = 0;
            }
            else if (dcf77_second != 59)   /* re-detected or out of sync */
                dcf77_status &=  ~(1<<DCF77_STATUS_INIT);   
            else
                dcf77_second   = 0;        /* this is second 1 */
        }
        else if (dcf77_second > 58)        /* missed startbit */
            dcf77_status &=  ~(1<<DCF77_STATUS_INIT);   
        
        dcf77_call_fall();
    }
    else {
        register u16 low_time;

        get_ticks(rise_time);              /* remember time of falling edge */

        low_time  = rise_time-fall_time;     

        /* int raised on rising edge */
        mcucr &= ~(1<<ISC10);              /* raise int on falling edge */
        if (low_time <= DCF77_ZERO)
            dcf77_bit_received(0);         /* 'zero' received */
        else if (low_time <= DCF77_ONE)
            dcf77_bit_received(1);         /* 'one' received */

        dcf77_call_rise();

    }
    outp(mcucr, MCUCR);
}


void dcf77_bit_received(u08 bit)
/* dcf77 bitstream decoder */
/* <bit>: last bit received (bit number = dcf77_second) */
/* decodes minute, hour, day, month and year from bitstream */
{
    static   u08 value;
    static   u08 parity;
    static   u08 startbit;
    register u08 b;


    b = bit&1;
    if (b) 
        dcf77_status |= (1<<DCF77_STATUS_BIT);
    else
        dcf77_status &= ~(1<<DCF77_STATUS_BIT);

    if (dcf77_status & (1<<DCF77_STATUS_INIT))
        switch (dcf77_second) {
            case 20: /* startbit */
                startbit = b; break;
            /* decode numbers */
            case 21: /* minute */
            case 29: /* hour */
            case 36: /* day */
                parity  = value = b; break;   
            case 42: /* weekday */
                dcf77_day = value;
                parity ^= b; value = b; break;   
            case 45: /* month */
                dcf77_wday = value;
                parity ^= b; value = b; break;   
            case 50: /* year */
                dcf77_month = value;
                parity ^= b; value = b; break;   
            case 22: /* minute */
            case 30: /* hour */
            case 37: /* day */
            case 43: /* weekday */
            case 46: /* month */
            case 51: /* year */
                parity ^= b; value += (b<<1); break;
            case 23: /* minute */
            case 31: /* hour */
            case 38: /* day */
            case 44: /* weekday */
            case 47: /* month */
            case 52: /* year */
                parity ^= b; value += (b<<2); break;
            case 24: /* minute */
            case 32: /* hour */
            case 39: /* day */
            case 48: /* month */
            case 53: /* year */
                parity ^= b; value += (b<<3); break;
            case 25: /* minute */
            case 33: /* hour */
            case 40: /* day */
            case 49: /* month */
            case 54: /* year */
                parity ^= b; if (b) value+=10; break;
            case 26: /* minute */
            case 34: /* hour */
            case 41: /* day */
            case 55: /* year */
                parity ^= b; if (b) value+=20; break;
            case 27: /* minute */
            case 56: /* year */
                parity ^= b; if (b) value+=40; break;
            case 57: /* year */
                parity ^= b; if (b) value+=80; break;
            
            case 28: /* minute */
                dcf77_minute = value;
                if (b!=parity || !startbit)
                    dcf77_status &= ~(1<<DCF77_STATUS_MIN_VALID);
                else
                    dcf77_status |= (1<<DCF77_STATUS_MIN_VALID);
                break;
            case 35: /* hour */
                dcf77_hour = value;
                if (b!=parity || !startbit)
                    dcf77_status &= ~(1<<DCF77_STATUS_HOUR_VALID);
                else
                    dcf77_status |= (1<<DCF77_STATUS_HOUR_VALID);
                break;
            case 58: /* day, month, year */
                dcf77_year = value;
                if (b!=parity || !startbit)
                    dcf77_status &= ~(1<<DCF77_STATUS_DATE_VALID);
                else
                    dcf77_status |= (1<<DCF77_STATUS_DATE_VALID);
                break;
        }
}
    


void dcf77_init(void)
{
    cbi(DDRD,  3);                         /* use PD3 for input */
    sbi(PORTD, 3);                         /* activate pullup on PD3 */

    outp((1<<ISC11), MCUCR);               /* raise int1 on falling edge */
    outp(0x80, GIMSK);                     /* enable external int1 */

    dcf77_call_fall = dcf77_dummy;
    dcf77_call_rise = dcf77_dummy;

    dcf77_status   = 0;                    /* startbit not detected yet */
    dcf77_start    = 0xfe; 
    dcf77_second   = 0xff;                 /* second counter */
}